-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.groestl_pkg.all;

-- Possible generics values:
-- hs = {HASH_SIZE_256, HASH_SIZE_512}

entity groestl_control_pq_parallel_ppl is
	generic (
		HS : integer:=HASH_SIZE_256;	   
		PPL : integer := 2;
		n : integer := GROESTL_DATA_SIZE_SMALL
	);
	port (
		clk						: in std_logic;
		rst						: in std_logic;	
		
		-- fifo signals
		src_ready				: in std_logic_vector(ppl-1 downto 0);
		src_read				: out std_logic_vector(ppl-1 downto 0);
		dst_ready 				: in std_logic;
		dst_write				: out std_logic;

		-- input			  
		smsg 					: out std_logic_vector(log2(ppl)-1 downto 0);
		schain					: out std_logic_vector(log2(ppl)-1 downto 0);
		ein						: out std_logic_vector(ppl-1 downto 0);	
		en_ctr					: out std_logic_vector(ppl-1 downto 0);
		en_len					: out std_logic_vector(ppl-1 downto 0);		
		en_c					: out std_logic_vector(ppl-1 downto 0);		
		load_id					: out std_logic_vector(ppl-1 downto 0);
		final_segment			: in std_logic_vector(ppl-1 downto 0);	
		c						: in std_logic_vector(ppl*16-1 downto 0);	
        
        -- pad
        comp_rem_e0             : in std_logic_vector(ppl-1 downto 0);
        comp_lb_e0              : in std_logic_vector(ppl-1 downto 0);
		comp_extra_block		: in std_logic_vector(ppl-1 downto 0);
        en_lb                   : out std_logic_vector(ppl-1 downto 0);
        clr_lb                  : out std_logic_vector(ppl-1 downto 0);
		sel_pad                 : out std_logic_vector(ppl-1 downto 0);
        rst_ctr                 : out std_logic_vector(ppl-1 downto 0);
		sel_out					: out std_logic_vector(ppl-1 downto 0); 
        spos                    : out std_logic_vector(PPL*2-1 downto 0);
		
		-- shared
		finalization			: out std_logic;
		init1					: out std_logic;
		init2					: out std_logic;
		init3					: out std_logic;	
		init4					: out std_logic;
		load_ctr				: out std_logic;
		wr_ctr					: out std_logic;
		wr_result				: out std_logic_vector(ppl-1 downto 0);
		wr_state				: out std_logic;

		-- output						  				 
		load_hash 				: out std_logic_vector(ppl-1 downto 0);
		shift_dout				: out std_logic_vector(ppl-1 downto 0);
		sout 					: out std_logic_vector(log2(ppl)-1 downto 0);
		sid 					: out std_logic
	);
end groestl_control_pq_parallel_ppl;

architecture struct of groestl_control_pq_parallel_ppl is
	-- fsm1
	signal block_ready_set, msg_end_set, load_next_block : std_logic_vector(ppl-1 downto 0);	
	-- fsm2
		-- fsm1 communications
	signal block_ready_clr, msg_end_clr : std_logic_vector(ppl-1 downto 0); --out
	signal block_ready, msg_end : std_logic_vector(ppl-1 downto 0); --in	   
	
		-- fsm2 communications
	signal output_busy_set : std_logic_vector(ppl-1 downto 0); --out
	signal output_busy : std_logic_vector(ppl-1 downto 0); --in  
	signal smsg_wire, schain_wire : std_logic_vector(log2(ppl)-1 downto 0);
	
	-- fsm3										
	signal output_busy_clr : std_logic_vector(ppl-1 downto 0); 
	signal shift_dout_fsm3, load_hash_fsm2 : std_logic_vector(ppl-1 downto 0);
			

	-- pipelined signals
	signal final_wire, init1_wire, init2_wire, init3_wire : std_logic;
	signal init4_wire : std_logic;
	signal load_ctr_wire, wr_ctr_wire : std_logic;
	signal wr_state_wire, lo_wire : std_logic;
	signal wr_result_wire : std_logic_vector(ppl-1 downto 0);

	-- output	
	signal sout_wire : std_logic_vector(log2(ppl)-1 downto 0);
	signal sid_wire  : std_logic;
	signal dst_write_wire : std_logic;
begin
    Fsm1Gen: for i in 0 to ppl-1 generate
        Fsm1Inst : entity work.groestl_fsm1(nocounter)
            generic map (mw=>n)
            port map (clk => clk, rst => rst, 
            c => c(PPL*16-(i+1)*16+15 downto PPL*16-(i+1)*16), en_len => en_len(i), en_ctr => en_ctr(i),
            ein => ein(i),load_next_block => load_next_block(i),
            block_ready_set => block_ready_set(i), msg_end_set => msg_end_set(i), load_id => load_id(i),
            src_ready => src_ready(i), src_read => src_read(i), final_segment=>final_segment(i), en_c => en_c(i),
            comp_rem_e0 => comp_rem_e0(i), comp_lb_e0 => comp_lb_e0(i), en_lb => en_lb(i), clr_lb => clr_lb(i), 
            comp_extra_block => comp_extra_block(i),  spos => spos((PPL-i)*2-1 downto (PPL-i-1)*2), sel_pad => sel_pad(i), rst_ctr => rst_ctr(i));

    end generate;

	FSM2Inst : entity work.groestl_fsm2(beh)
		generic	map (hs=>hs, PPL=>PPL)
		port map (
            clk => clk, 
            rst => rst, 
            msg_end_clr => msg_end_clr,
            msg_end => msg_end,             
            block_ready => block_ready, 
            block_ready_clr => block_ready_clr, 
            output_busy => output_busy,
            output_busy_set => output_busy_set,
            load_hash => load_hash_fsm2, 
			smsg => smsg_wire,	
			schain => schain_wire,
			
            -- shared control
            final => final_wire, 
            init1 => init1_wire, 
            init2 => init2_wire, 
            init3 => init3_wire,
			init4 => init4_wire,
            load_ctr => load_ctr_wire,
            wr_ctr => wr_ctr_wire, 
            wr_result => wr_result_wire, 
            wr_state => wr_state_wire		
		);

	FSM3Inst : entity work.sha3_fsm3_ppl(beh)
		generic map ( h => HS, w=> w, ppl => ppl )
		port map (
		clk => clk, rst => rst, 
		shift_dout => shift_dout_fsm3, sout => sout_wire,  sid => sid_wire,
		output_busy => output_busy, output_busy_clr => output_busy_clr,
		dst_ready => dst_ready, dst_write => dst_write_wire
	);	 

	ctrl_reg: process( clk )
	begin
		if rising_edge( clk ) then
			finalization <= final_wire;
			init1 <= init1_wire;
			init2 <= init2_wire;
			init3 <= init3_wire;
			init4 <= init4_wire;
			load_ctr <= load_ctr_wire;
			wr_ctr <= wr_ctr_wire;
			wr_result <= wr_result_wire;
			wr_state <= wr_state_wire;	 
			schain <= schain_wire;
			
            -- output
            smsg <= smsg_wire;
			sout <= sout_wire;
			sid <= sid_wire;			
			shift_dout <= shift_dout_fsm3 or load_hash_fsm2; 
			load_hash  <= load_hash_fsm2;	
			dst_write <= dst_write_wire;	
		end if;
	end process;
    
    
	CommGen :  for i in 0 to ppl-1 generate
		load_next_block(i) <= not block_ready(i);
		sr_blk_ready : sr_reg 
			port map ( rst => rst, clk => clk, set => block_ready_set(i), clr => block_ready_clr(i), output => block_ready(i));	
		sr_msg_end : sr_reg 
			port map ( rst => rst, clk => clk, set => msg_end_set(i), clr => msg_end_clr(i), output => msg_end(i));	
		sr_output_write : sr_reg 
			port map ( rst => rst, clk => clk, set => output_busy_set(i), clr => output_busy_clr(i), output => output_busy(i));			
	end generate; 
end struct;