-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;  
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;

entity fifo_ppl is
	generic (
		depth 			: integer := 512;	
		n 				: integer := 64);
	port (
		clk				: in std_logic;
		rst				: in std_logic;
		write			: in std_logic; 
		read			: in std_logic;
		din 			: in std_logic_vector(n-1 downto 0);
		dout	 		: out std_logic_vector(n-1 downto 0);
		full			: out std_logic; 
		empty 			: out std_logic);
end fifo_ppl;

architecture struct of fifo_ppl is	
	constant log2depth 		: integer := log2(depth);

	type 	mem is array (0 to depth-1) of std_logic_vector(n-1 downto 0);
	signal 	memory 		: mem;

	attribute ram_style : string ;
	attribute ram_style of memory : signal is "block";	   					
		
	signal readpointer, readaddr  	: std_logic_vector(log2depth-1 downto 0);
	signal writepointer 	: std_logic_vector(log2depth-1 downto 0);
	signal bytecounter  	: std_logic_vector(log2depth downto 0);
		
	signal read_en, write_en, byte_en, load_output : std_logic;
	signal byte_add, read_fifo : std_logic;
	
	signal bctr_in : std_logic_vector(log2depth downto 0);
	
	signal pre_dout : std_logic_vector(n-1 downto 0);
	
	type fifo_state_type is ( no_data, data_at_bram, data_ready );
	signal ns, cs :fifo_state_type; -- current state, next state
	
	signal zbyte1, zbyte2 : std_logic;
begin		 
	fiforam : process(clk)
	begin
		if rising_edge(clk) then
			if (write_en = '1') then
				memory(conv_integer(writepointer)) <= din;
			end if;	 	
			if ( read_fifo = '1' ) then
				readaddr <= readpointer;
			end if;
		end if;
	end process; 		
	pre_dout <= memory(conv_integer(readaddr));
	
	bctr_in <= bytecounter + 1 when (byte_add = '1') else bytecounter - 1;	
	cntrs_gen : process(clk)
	begin
		if( rising_edge(clk) ) then	
			if rst = '1' then				
				bytecounter  <= (others => '0');  --differences (write pointer - read pointer)
				readpointer  <= (others => '0');
				writepointer <= (others => '0'); 
			else
				if ( byte_en = '1' ) then
					bytecounter <= bctr_in;
				end if;
				if ( write_en = '1' ) then
					writepointer <= writepointer + 1;
				end if;
				if ( read_en = '1' ) then
					readpointer <= readpointer + 1;	
				end if;				
			end if;
		end if;
	end process;   
	
	
	state_transition: process ( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' then
				cs <= no_data;
			else
				cs <= ns;
			end if;
		end if;
	end process;
	
	output_proc: process ( cs, zbyte2, zbyte1, read, write  )
	begin
		read_en <= '0'; write_en <= '0'; byte_en <= '0';
		load_output <= '0';	read_fifo <= '0'; byte_add <= '0';
		if cs = no_data then
			-- At this point, we don't have data at any of the stage;
			empty <= '1';
			if ( write = '1' ) then
				byte_add <= '1';
				byte_en <= '1';
				write_en <= '1';
				read_fifo <= '1';
				read_en <= '1';
				-- //////////
				-- Next state
				ns <= data_at_bram;
			end if;					
		elsif cs = data_at_bram then 
			
			-- Comment: 
			-- 	if data is at bram, we want to load the data onto the output register 
			--	so that it becomes ready in the next clock cycle
			empty <= '1';
			load_output <= '1';
			read_fifo <= '1';
			read_en <= '1';
			if ( write = '1' ) then				
				byte_add <= '1';
				byte_en <= '1';
				write_en <= '1';
			end if;
			
			-- //////////
			-- Next State
			ns <= data_ready;
		elsif cs = data_ready then
			empty <= '0';			
			if ( write = '1' and read = '0' ) then -- write only
				write_en <= '1';				
				byte_add <= '1';
				byte_en <= '1';
			elsif ( write = '1' and read = '1') then -- read and write at the same time
				write_en <= '1';	 
				read_fifo <= '1';
				load_output <= '1'; -- load the new address
				read_en <= '1';				
				
				-- //////////
				-- Next State
				if (zbyte1 = '1') then 
					ns <= data_at_bram;
				else
					ns <= data_ready;
				end if;
			elsif ( write = '0' and read = '1' ) then -- read only
				load_output <= '1';
				byte_en <= '1';	 
				read_fifo <= '1';
				
				if ( zbyte1 = '0' and zbyte2 = '0')  then
					read_en <= '1';
				end if;
				
				-- //////////
				-- Next State
				if ( zbyte1 = '1' ) then 
					ns <= no_data;
				else
					ns <= data_ready;
				end if;
			end if;
		end if;
	end process;
				
				
	-- Comparator 		
	zbyte1 <= '1' when bytecounter(log2depth-1 downto 0) = 1 else '0';
	zbyte2 <= '1' when bytecounter(log2depth-1 downto 0) = 2 else '0';	
	
	full <= bytecounter(log2depth);		
	
	-- register the output signals
	output_reg : process( clk )
	begin
		if rising_edge( clk ) then
			if ( load_output = '1' ) then
				dout <= pre_dout;
			end if;	
		end if;
	end process;	
end struct;

			 